/* 
 * Copyright 2015-2016 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SCHEDULER_APP_H
#define SCHEDULER_APP_H

#include <QProcess>
#include <QSet>
#include <QFile>
#include <QTimer>
#include <QHash>
#include "common.h"

QT_BEGIN_NAMESPACE
class QLocalSocket;
class QLocalServer;
class QNetworkConfigurationManager;
class QLockFile;
QT_END_NAMESPACE

#define APPNAME "spoofer-scheduler"

class SubnetAddr : public QPair<QHostAddress,int> {
public:
    SubnetAddr(const QHostAddress &haddr, int pfxlen) : QPair(haddr, pfxlen) {}
    SubnetAddr(const QPair<QHostAddress,int> &pair) : QPair(pair) {}
    QHostAddress addr() const { return first; }
    int pfxlen() const { return second; }
    SubnetAddr prefix() const;
    QString toString() const {
	return addr().toString() % QSL("/") % QString::number(pfxlen());
    }
};

class App : public QCoreApplication, public SpooferBase {
    Q_OBJECT

protected:
    class AppLog : public QFile {
    public:
	static QString makeName();
	AppLog() : QFile(makeName()) {}
    };

    struct RunRecord {
	time_t t;
	unsigned int errors;
    };

    App(int &argc, char **argv) :
	QCoreApplication(argc, argv), SpooferBase(),
	optShareData(-1), optInitOnly(false),
	optDumpPaths(false), optDetach(false),
	optStandalone(false), optLogfile(false),
	isInteractive(false), isService(false), paused(false),
	lockfile(), privServer(), unprivServer(), uiSet(), prober(),
	scheduledAddrs(),
	nextProberStart(), proberTimer(), netPollTimer(), pastRuns()
	{ }
public:
    virtual ~App();
    static App *newapp(int &argc, char **argv);
    void cleanup(void);
    virtual bool parseCommandLine(QCommandLineParser &clp);
    virtual void dumpPaths(void) const;
    virtual bool chooseDataDir(void);
    virtual bool init(int &exitCode);
    virtual bool prestart(int &exitCode) { Q_UNUSED(exitCode); return true; }
    virtual void readyService(int &exitCode) const { Q_UNUSED(exitCode); }
    virtual void endService(int &exitCode) { Q_UNUSED(exitCode); }
    virtual void end() const {};
    virtual bool initSignals() = 0;
    virtual void pause();
    virtual void resume();
    static const QString name;
    static int argc;
    static char **argv;
    int optShareData;
    bool optInitOnly;
    bool optDumpPaths;
    bool optDetach;
    bool optStandalone;
    bool optLogfile;
    bool isInteractive;
    bool isService;
    bool paused;
protected:
    static QString dataDir;

    QLockFile *lockfile;
    QLocalServer *privServer;
    QLocalServer *unprivServer;
    QSet<QLocalSocket *> uiSet;
    QProcess *prober;
    QString proberOutputFileName;
    QList<SubnetAddr> scheduledAddrs;
    sc_msg_scheduled nextProberStart;
    QTimer proberTimer;
    QTimer netPollTimer;
    // pastRuns is indexed not by the full address of interfaces, but by their
    // subnet prefixes, so that:  1) a new prober run is not triggered when
    // the host moves within a subnet or uses a different temporary IPv6
    // address; 2) history isn't cluttered with random temporary IPv6 addrs.
    QHash<SubnetAddr, RunRecord> pastRuns;
    QLocalServer *listen(bool privileged);

    bool opAllowed(QLocalSocket *ui, const Config::MemberBase &cfgItem);
    bool opAllowedVerbose(QLocalSocket *ui, const Config::MemberBase &cfgItem);
    void recordRun(bool success);
    virtual void startProber();
    virtual void killProber();
    virtual void shutdown() { this->exit(0); }
    QList<SubnetAddr> getAddresses() const;
    QNetworkConfigurationManager *ncm; // experimental

private slots:
    void uiRead();
    void uiAccept();
    void uiDelete();
    void proberStarted();
    void proberError(QProcess::ProcessError);
    void proberFinished(int exitCode, QProcess::ExitStatus exitStatus);
    void dumpNetCfg();
    void handleNetChange();
    void scheduleNextProber();
};

#endif // SCHEDULER_APP_H
